/* This filter uses Olli's SoundTouch lib for changing
   tempo/pitch. */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "af.h"

// Initialization and runtime control
static int control(struct af_instance_s* af, int cmd, void* arg)
{
  void *STObj = af->setup;
  switch(cmd){
  case AF_CONTROL_REINIT:{
    af_data_t* 	   n   = (af_data_t*)arg; // New configureation
    SoundTouchReinit(STObj,n->len, 
		     n->rate, n->nch, n->format, n->bps,
		     &(af->mul.n), &(af->mul.d));
    // The af->delay isn't really importent for now, as it is not used. 
    // Also, this plugin delay is dynamic and not so accurate
    memcpy(af->data,(af_data_t*)arg,sizeof(af_data_t));
    af_msg(AF_MSG_VERBOSE,"[SoundTouch] Was reinitialized, rate=%iHz, nch = %i, format = 0x%08X and bps = %i\n",af->data->rate,af->data->nch,af->data->format,af->data->bps);
    return AF_OK;
  }
  case AF_CONTROL_TEMPO | AF_CONTROL_SET:{
    // No need for reinint!
    if(SoundTouchTempo(af->setup, ((double*)arg)[0], &(af->data->len), &(af->mul.n), &(af->mul.d))!=1) {
      af_msg(AF_MSG_ERROR, "[SoundTouch] Setting Tempo (%f) failed!\n", ((float*)arg)[0]);
      return AF_ERROR;
    }
    return AF_OK;
  }
  case AF_CONTROL_PITCH | AF_CONTROL_SET:{
    // No need for reinint!
    if(SoundTouchPitch(af->setup, ((float*)arg)[0], &(af->data->len), &(af->mul.n), &(af->mul.d))!=1) {
      af_msg(AF_MSG_ERROR, "[SoundTouch] Setting Pitch (%f) failed!\n", ((float*)arg)[0]);
      return AF_ERROR;
    }
    return AF_OK;
  }

  case AF_CONTROL_ST_RATE | AF_CONTROL_SET:{
    // No need for reinint!
    if(SoundTouchRate(af->setup, ((float*)arg)[0], &(af->data->len), &(af->mul.n), &(af->mul.d))!=1) {
      af_msg(AF_MSG_ERROR, "[SoundTouch] Setting Pitch (%f) failed!\n", ((float*)arg)[0]);
      return AF_ERROR;
    }
    return AF_OK;
  }
  }
  return AF_UNKNOWN;
}

// Deallocate memory 
static void uninit(struct af_instance_s* af)
{
  void *        STObj = af->setup;
  DestroySoundTouchObj(STObj);
  if(af->data)
    free(af->data);
}

// Filter data through filter
static af_data_t* play(struct af_instance_s* af, af_data_t* data)
{

  af_data_t*    c     = data;	 // Current working data
  af_data_t*    l     = af->data; // Local data
  void *        STObj = af->setup;
  int           len   = l->len;

  if(AF_OK != RESIZE_LOCAL_BUFFER(af,data))
    return NULL;

  SoundTouchProcess(STObj,
		    c->audio, &(c->len), c->bps, c->nch,
		    l->audio, &(len));

  // Set output data
  c->audio = l->audio;
  c->len   = len;
  c->rate  = l->rate;
  
  return c;
}

// Allocate memory and set function pointers
static int open(af_instance_t* af){
  af->control=control;
  af->uninit=uninit;
  af->play=play;
  af->mul.d=1;
  af->mul.n=1;
  af->data=malloc(sizeof(af_data_t));
  af->setup=MakeNewSoundTouchObj();
  if(af->data == NULL||af->setup == NULL)
    return AF_ERROR;
  return AF_OK;
}

// Description of this filter
af_info_t af_info_stouch = {
    "Sound-Touch libs bindings (the C part)",
    "stouch",
    "Shachar Raindel (Based on Olli Parviainen's Sound-Touch libs)",
    "Mainly for smoother playspeed change",
    AF_FLAGS_REENTRANT,
    open
};
