//This is a heavyly modified version of the SoundStretch main file
//Which acts as binding between the cpp library and the c player core
/******************************************************************************
 *
 * The SoundStretch main routine. This program processes WAV audio files by
 * using the SoundTouch library routines.
 *
 * Author        : Copyright (c) Olli Parviainen 2002
 * Author e-mail : oparviai@iki.fi
 * File created  : 13-Jan-2002
 * Last modified : 21-Feb-2002
 *
 * Change History :
 *
 * 21-Feb-2002 OP : Changed version number to v1.01 due to changes in
 *                  WavFile.cpp
 *
 * License :
 *
 *  This file is part of SoundTouch sound processing library.
 *
 *  SoundTouch is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  SoundTouch is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with SoundTouch; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdexcept>
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include "SoundTouch.h"

using namespace std;

#ifndef min
#define min(a,b)(((a)>(b))?(b):(a))
#endif

#ifndef max
#define max(a,b)(((a)>(b))?(a):(b))
#endif

extern "C" {

#define BUFF_SIZE       2048
#define BUFF_SIZE_BYTES (2 * BUFF_SIZE)

#define MINIMAL_STEP_MULTI 1000 

#define MAGIC_STPlugD 0x137611ec

struct STPlugD {
  int magic;
  STPlugD():Tempo(1.0),Pitch(1.0),Rate(1.0) {
    magic = MAGIC_STPlugD;
  }
  class SoundTouch Proccessor;
  float Tempo, Pitch, Rate;
};
  
static const char helloText[] = 
    "\n"
    "   SoundStretch v1.01 -  Written by Olli Parviainen 2001 - 2002\n"
    "Mplayer bind by Shachar Raindel<shacharrATt2.technion.ac.il>2003\n"
    "==================================================================\n"
    "author e-mail: <oparviai@iki.fi> - WWW: http://www.iki.fi/oparviai\n"
    "\n";

#if 0
static void openFiles(WavInFile **inFile, WavOutFile **outFile, const RunParameters *params)
{
    int bits, samplerate, channels;

    // open input file...
    *inFile = new WavInFile(params->inFileName);

    // ... open output file with same sound parameters
    bits = (*inFile)->getNumBits();
    samplerate = (*inFile)->getSampleRate();
    channels = (*inFile)->getNumChannels();

    *outFile = new WavOutFile(params->outFileName, samplerate, bits, channels);
}
#endif

long int lround(double x)
{
  if(x>0) return (x+0.5);
  return (x-0.5);
}

int ggt(int a,int b)
{
    if (b==0) { return(a); }
    return(ggt(b, a % b));
}

double ggtd(double a, double b)
{
  if ((a<32768) && (b<32768)) {
		return((double)ggt((int)lround(a),(int)lround(b)));
	} else {
	    	if ((int)b<0.5) {
		    return(a);
		} else {
		    return(ggtd(b,a-b*floor(a/b)));
		}
	}
}

void kr(double *a,double *b)
{
	double t;

	t=ggtd(*a,*b);
	*a=*a/t;
	*b=*b/t;
}

void emb(double q,double *a,double *b,double eps)
{
    double q0,qabs,qvor,qnk,am,amm,bmm,bm;
    int i;

    qabs=fabs(q);
    q0=qabs;
    *a=1.0;
    *b=0.0;
    am=0.0;
    bm=1.0;

    i=0;

    do {
	i++;
	if (i>30) break;
	qvor=floor(q0+eps);
	qnk=q0-qvor;

	amm=am;
	bmm=bm;
	am=*a;
	bm=*b;

	*a=qvor*am+amm;
	*b=qvor*bm+bmm;
	kr(a,b);

	if (qnk<eps) break;

	q0=1/qnk;
    } while (fabs(qabs-(*a)/(*b))>=eps);

    if (fabs(qabs-(*a)/(*b))>eps) return;
    if (q<0) *a=-*a;
}



void CalcMul(struct STPlugD *Obj,int *mul_n, int *mul_d, int buff_len)
{

    double a, b, f= Obj->Tempo * Obj->Rate;
    int r;

    emb(f, &a, &b, 1e-7);


    *mul_n=(int)a;
    *mul_d=(int)b;
#if 0
  int i=1,j=1,d;
  float target= Obj->Tempo * Obj->Rate;
  if(buff_len < 1000) buff_len = 1000; /*Force some minimal accuracy*/
  while((buff_len*fabs(((float)i)/((float)j) - target))> 100) {
    if(((float)i)/((float)j) > target) {
      j++;
    }else {
      i++;
    }
  }
  d = gcd(i,j);
  *mul_n = i/d;
  *mul_d = j/d;
#endif
}

void SoundTouchReinit(void *STObj, int buff_len,
		      int rate, int nch, int format, int bps,
		      int *mul_n, int *mul_d)
{
  struct STPlugD *Obj = (struct STPlugD *)STObj;
  if(Obj == NULL||Obj->magic != MAGIC_STPlugD) return; //Someone missbehaved. Ignore him.
  CalcMul(Obj,mul_n,mul_d,buff_len);
  SoundTouch *pSoundTouch = &(Obj->Proccessor);
  pSoundTouch->setSampleRate(rate);
  pSoundTouch->setChannels(nch);
  /* Bug: FORMAT handeling!*/
  
  
}

void DestroySoundTouchObj(void *STObj)
{
  struct STPlugD *Obj = (struct STPlugD *)STObj;
  if(Obj == NULL||Obj->magic != MAGIC_STPlugD) return; //Someone missbehaved. Ignore him.
  delete Obj;
}

int SoundTouchTempo(void *STObj, double tempo, int buff_len, int *mul_n, int *mul_d)
{
  try {
    struct STPlugD *Obj = (struct STPlugD *)STObj;
    if(Obj == NULL||Obj->magic != MAGIC_STPlugD) return -1; //Someone missbehaved. Ignore him.
    SoundTouch *pSoundTouch = &(Obj->Proccessor);
#ifdef MINIMAL_STEP_MULTI
    tempo = ((double)lround(tempo*((double)MINIMAL_STEP_MULTI)))/((double)MINIMAL_STEP_MULTI);
#endif
    pSoundTouch->setTempo(tempo);
    Obj->Tempo = tempo;
    CalcMul(Obj,mul_n,mul_d,buff_len);
    return 1;
  }catch (runtime_error &e) {
    return -1;
  }
}

int SoundTouchPitch(void *STObj, double pitch, int buff_len, int *mul_n, int *mul_d)
{
  try {
    struct STPlugD *Obj = (struct STPlugD *)STObj;
    if(Obj == NULL||Obj->magic != MAGIC_STPlugD) return -1; //Someone missbehaved. Ignore him.
    SoundTouch *pSoundTouch = &(Obj->Proccessor);
#ifdef MINIMAL_STEP_MULTI
    pitch = ((double)lround(pitch*((double)MINIMAL_STEP_MULTI)))/((double)MINIMAL_STEP_MULTI);
#endif
    pSoundTouch->setPitch(pitch);
    Obj->Pitch = pitch;
    CalcMul(Obj,mul_n,mul_d,buff_len);
    return 1;
  }catch (runtime_error &e) {
    return -1;
  }
}


int SoundTouchRate(void *STObj, double rate, int buff_len, int *mul_n, int *mul_d)
{
  try {
    struct STPlugD *Obj = (struct STPlugD *)STObj;
    if(Obj == NULL||Obj->magic != MAGIC_STPlugD) return -1; //Someone missbehaved. Ignore him.
    SoundTouch *pSoundTouch = &(Obj->Proccessor);
#ifdef MINIMAL_STEP_MULTI
    rate = ((double)lround(rate*((double)MINIMAL_STEP_MULTI)))/((double)MINIMAL_STEP_MULTI);
#endif
    pSoundTouch->setRate(rate);
    Obj->Rate = rate;
    CalcMul(Obj,mul_n,mul_d,buff_len);
    return 1;
  }catch (runtime_error &e) {
    return -1;
  }
}


#if 0
// Sets the 'SoundTouch' object up according to input file sound format & 
// command line parameters
void setup(SoundTouch *pSoundTouch, const WavInFile *inFile, const RunP
{
  int sampleRate;
  int channels;
  
  sampleRate = inFile->getSampleRate();
  channels = inFile->getNumChannels();
  pSoundTouch->setSampleRate(sampleRate);
  pSoundTouch->setChannels(channels);

  pSoundTouch->setTempoChange(params->tempoDelta);
  pSoundTouch->setPitchSemiTones(params->pitchDelta);
  pSoundTouch->setRateChange(params->rateDelta);

  pSoundTouch->setSetting(SETTING_USE_QUICKSEEK, params->quick);
  pSoundTouch->setSetting(SETTING_USE_AA_FILTER, !params->noAntiAlias);

    // print processing information
  printf("Processing the file with the following changes:\n");
  printf("  tempo change = %+g %%\n", params->tempoDelta);
  printf("  pitch change = %+g semitones\n", params->pitchDelta);
  printf("  rate change  = %+g %%\n\n", params->rateDelta);
  printf("Working...");
}

#endif


// Processes the sound: Read sound data from the input file ('inFile'), 
// process with the processing object ('pSoundTouch') and write to the 
// output file ('outFile').
void SoundTouchProcess(void *STObj, 
		       void *InputData, int *ilen, int bps, int nch,
		       void *OutputData, int *olen)
{
  int nBytes;
  int nSamples;
  int bytesPerSample;
  int buffSizeSamples;
  int obuffSizeSamples;
  
  struct STPlugD *Obj = (struct STPlugD *)STObj;
  if(Obj == NULL||Obj->magic != MAGIC_STPlugD) return; //Someone missbehaved. Ignore him.

  SoundTouch *pSoundTouch = &(Obj->Proccessor);
  // Calculate number of samples in the buffer
  bytesPerSample = bps*nch;
  buffSizeSamples = *ilen / bytesPerSample;
  obuffSizeSamples = *olen/ bytesPerSample;
  // Feed the sound data into the processing object
  pSoundTouch->putSamples((short int *)InputData, buffSizeSamples);
  *ilen = 0; // We inserted all of the samples into the pipeline
  *olen = 0;

  // Receive processed data from the processing object and write into
  // the output buffer. Iterate this in loop until all the available sound
  // data is extracted from the processing object, or until the output buffer
  // is full
  do {
    // receive processed audio data
    nSamples = pSoundTouch->receiveSamples((short int *)OutputData, obuffSizeSamples);
    // Reduce it from the available space
    ((short int *)OutputData) += nSamples;
    obuffSizeSamples -= nSamples;
    *olen += (nSamples *  bytesPerSample);
  } while((nSamples != 0) && (obuffSizeSamples > 0));

#if 0
    // Extract few last samples from the processing pipeline.
    // First call the 'flush' function.
    pSoundTouch->flush();
    // ...then iterate the receiveSamples loop until all available data from
    // the processing object is extracted and written to the output file.
    do {
        nSamples = pSoundTouch->receiveSamples(buffer, buffSizeSamples);
        nBytes = nSamples * bytesPerSample;
        outFile->write(buffer, nBytes);
    } while (nSamples != 0);
#endif
}

void *MakeNewSoundTouchObj()
{
  try {
    struct STPlugD *NewObject;
    NewObject = new struct STPlugD;
    return (void *) NewObject;
  }catch (runtime_error &e)  {
    return NULL;
  }
}


#if 0
int main(const int nParams, const char *paramStr[])
{
    WavInFile *inFile;
    WavOutFile *outFile;
    SoundTouch SoundTouch;
    RunParameters *params;

    printf(helloText);
    try {
        // Parse command line parameters
        params = new RunParameters(nParams, paramStr);

        // Open input & output files
        openFiles(&inFile, &outFile, params);

        // Setup the 'SoundTouch' object for processing the sound
        setup(&SoundTouch, inFile, params);

        // Process the sound
        process(&SoundTouch, inFile, outFile);

        // Close WAV file handles & dispose of the objects
        delete inFile;
        delete outFile;
        delete params;

        printf("Done!\n");
    } catch (runtime_error &e) {
        // An exception occurred during processing, display an error message
        printf("%s\n", e.what());
        return -1;
    }

    return 0;
}

#endif
}
